"""
This module includes utility functions that are used in the project.
"""
import pandas as pd
from typing import List, Literal, Optional, Tuple
import numpy as np
import pandas_market_calendars as mcal

DEF_CALENDAR = 'NYSE'


def get_trading_days(start: str | pd.Timestamp, end: str | pd.Timestamp,
                     calendar: str = DEF_CALENDAR) -> List[pd.Timestamp]:
    """ Function that returns the trading days between two dates. """
    if isinstance(start, str):
        start = pd.Timestamp(start)
    if isinstance(end, str):
        end = pd.Timestamp(end)

    nyse = mcal.get_calendar(calendar)
    return nyse.valid_days(start_date=start, end_date=end, tz=None)


def roll_day_to_bday(date: pd.Timestamp | str, offset: int = 0,
                     direction: Literal["forward", "backward", "f", "b"] = "forward",
                     calendar: str = DEF_CALENDAR) -> pd.Timestamp:
    """
    Function that roll a date to the next or previous business day, given an offset and a direction.
    With offset=0, the function returns the nearest business day to the input date.
    """
    if isinstance(date, str):
        date = pd.Timestamp(date)
    if offset < 0:
        raise ValueError("Offset must be a positive integer or zero (for negative offset use the direction parameter)")

    bday_offset = offset + 2 * int(offset / 5) + int(20 * offset / 252) + 5  # Adjust for weekends and holidays

    if direction in ["forward", "f"]:
        return get_trading_days(date, date + pd.Timedelta(days=bday_offset), calendar=calendar)[offset]

    else:
        return get_trading_days(date + pd.Timedelta(days=-bday_offset), date, calendar=calendar)[-(offset + 1)]


def compute_rtn(prices_df: pd.DataFrame) -> pd.DataFrame:
    """
    Compute the daily log returns from the prices DataFrame.
    """
    daily_returns = np.log(prices_df / prices_df.shift(1)).dropna(how='all')
    if isinstance(daily_returns, pd.Series):
        daily_returns = daily_returns.to_frame(name=daily_returns.name)

    return daily_returns.dropna(how='all')


def compute_cum_rtn(returns_df: pd.DataFrame,
                    period: Optional[Tuple[str, Optional[str]] | Literal["M", "3M", "1Y", "5Y", "YTD"] | int] = None
                    ) -> pd.DataFrame:
    """
    Function that computes the cumulative returns for a given period, which correspond to the max lookback date
    (except for the tuple case).

    Period can be:
    - A tuple of two explict dates (start, end) in the format 'YYYY-MM-DD',
    - A string representing the period ('3M', '1Y', '5Y', 'YTD'),
    - An integer representing the number of bdays.
    - None, which means the entire DataFrame.
    """
    if period is None:
        returns_data = returns_df
    elif isinstance(period, tuple):
        start, end = period

        if end is None:  # If end is None, we take the last date in the DataFrame
            end = returns_df.index[-1]
        else:
            end = pd.Timestamp(end)
        start = pd.Timestamp(start)

        if start >= end or start not in returns_df.index or end not in returns_df.index:
            raise ValueError("Invalid dates.")
        returns_data = returns_df.loc[start:end]

    elif isinstance(period, str):
        if period == "YTD":
            returns_data = returns_df.loc[returns_df.index.year == returns_df.index[-1].year]
        else:
            returns_data = returns_df.loc[-string_period_to_bdays(period):]
    elif isinstance(period, int):
        returns_data = returns_df[-period:]
    else:
        raise ValueError("Invalid period type.")

    cum_returns = returns_data.cumsum()
    return cum_returns


def string_period_to_bdays(period: str) -> int | None:
    """Function that convert a string period to the number of bdays."""
    match period:
        case "M" | "1M":
            return 21
        case "3M":
            return 63
        case "1Y":
            return 252
        case "5Y":
            return 1260
        case _:
            return None



"""
This module manages the stock prices repository, which is a csv file containing the prices of the stocks, that will be
used during multiple executions of the program to avoid downloading the prices every time.
"""
import os
from typing import List, Optional, Literal

import pandas as pd
import yfinance as yf

from .utils import get_trading_days, roll_day_to_bday


class PriceRepo:
    """
    Class that manages the stock prices repository, which is a DataFrame/csv file containing the prices of the stocks,
    that will be used during multiple executions of the program to avoid downloading the prices every time.
    """

    def __init__(self, path: str | os.PathLike,
                 price_type: Literal['Open', 'High', 'Low', 'Close', 'Adj Close', 'Volume'] = 'Adj Close'):
        self.path = path
        if not os.path.exists(path):  # Create a dummy repo if it does not exist
            dummy_stock_prices = pd.DataFrame(index=pd.date_range(pd.Timestamp.today().date() - pd.DateOffset(days=1),
                                                                  pd.Timestamp.today().date(), freq='D', name='date'))
            dummy_stock_prices.to_csv(path)
        self.price_type = price_type
        self._price_df = pd.read_csv(path, index_col='date', parse_dates=True)

    @property
    def price_df(self) -> pd.DataFrame:
        return self._price_df

    def _download_stocks_prices(self, tickers_list: List[str], start: pd.Timestamp, end: pd.Timestamp) -> pd.DataFrame:
        """
        Download daily prices from yfinance from a list of tickers, within a specified date range.
        """
        prices_df = yf.download(tickers_list, start=start, end=end)[self.price_type]
        if isinstance(prices_df, pd.Series):  # If only one stock is downloaded, we convert the series to a DataFrame
            prices_df = prices_df.to_frame(name=tickers_list[0])
        return prices_df

    def _stock_in_repo(self, ticker: str, start: pd.Timestamp, end: pd.Timestamp) -> bool:
        """
        Check if the stock prices are already in the repo or if there are missing date (in the selected range) in the repo.
        """
        if ticker not in self._price_df.columns:
            return False
        if self._price_df.loc[start:end, ticker].isna().any():
            return False
        return True

    def get_prices(self, tickers_list: List[str], start: str | pd.Timestamp, end: Optional[str | pd.Timestamp] = None,
                   ) -> pd.DataFrame:
        """Function that return the df prices of a list of tickers, by loading them from the repo or by
        downloading them from yfinance if a stock prices is not available, then it returns the selected stocks
        and the update price_repo.

        Notes: The start and end must be in the format 'YYYY-MM-DD'. (if end is None,it will download until 2 weeks ago)
        """
        # Convert the dates to trading dates
        if end is None:
            end = (pd.Timestamp.today() - pd.Timedelta(days=13)).strftime('%Y-%m-%d')
        else:
            end = roll_day_to_bday(end, offset=1).strftime('%Y-%m-%d')  # Add one day to include the end date

        trading_period = pd.Index(get_trading_days(start, end))
        start, end, end_aft = trading_period[0], trading_period[-2], trading_period[-1]  # end_aft is one bday after
        trading_period = trading_period[:-1]  # Remove the extra day to avoid including it in the repo

        if start > end:
            raise ValueError("The start date must be before the end date.")

        if not trading_period.isin(self._price_df.index).all():  # if the period is not in the repo, reindex the repo
            index_name = self._price_df.index.name
            self._price_df = self._price_df.reindex(self._price_df.index.union(trading_period))
            self._price_df.index.name = index_name

        # Check which stocks need to be downloaded
        tickers_to_download = [ticker for ticker in tickers_list if not self._stock_in_repo(ticker, start, end)]

        if len(tickers_to_download) > 0:  # If all the stocks are already in the repo, return the prices selected
            new_prices = self._download_stocks_prices(tickers_to_download, start,
                                                      end_aft)  # end_aft to include the last day
            self._price_df.loc[start:end, tickers_to_download] = new_prices  # Update the repo with the new prices

            self._price_df.to_csv(self.path)  # Save the updated repo

        return self._price_df.loc[start:end, tickers_list]
